@@ -1,5 +1,5 @@
 /* GAIL - The GNOME Accessibility Implementation Library
- * Copyright 2001 Sun Microsystems Inc.
+ * Copyright 2001, 2002, 2003 Sun Microsystems Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -31,12 +31,15 @@ static gint                  gail_button_get_n_children   (AtkObject       *obj)
 static AtkObject*            gail_button_ref_child        (AtkObject       *obj,
                                                            gint            i);
 static AtkStateSet*          gail_button_ref_state_set    (AtkObject       *obj);
-static void                  gail_button_real_notify_gtk  (GObject         *obj,
-                                                           GParamSpec      *pspec);
+static void                  gail_button_notify_label_gtk (GObject         *obj,
+                                                           GParamSpec      *pspec,
+                                                           gpointer        data);
 
 static void                  gail_button_real_initialize  (AtkObject       *obj,
                                                            gpointer        data);
 static void                  gail_button_finalize         (GObject        *object);
+static void                  gail_button_init_textutil    (GailButton     *button,
+                                                           GtkWidget      *label);
 
 static void                  gail_button_pressed_enter_handler  (GtkWidget       *widget);
 static void                  gail_button_released_leave_handler (GtkWidget       *widget);
@@ -117,7 +120,10 @@ static AtkAttributeSet* gail_button_get_run_attributes
 static AtkAttributeSet* gail_button_get_default_attributes
                                                    (AtkText           *text);
 static GtkImage*             get_image_from_button      (GtkWidget      *button);
-static GtkWidget*            get_label_from_button      (GtkWidget      *button);
+static GtkWidget*            get_label_from_button      (GtkWidget      *button,
+                                                         gint           index,
+                                                         gboolean       allow_many);
+static gint                  get_n_labels_from_button   (GtkWidget      *button);
 static void                  set_role_for_button        (AtkObject      *accessible,
                                                          GtkWidget      *button);
 
@@ -198,8 +204,6 @@ gail_button_class_init (GailButtonClass *klass)
   class->ref_child = gail_button_ref_child;
   class->ref_state_set = gail_button_ref_state_set;
   class->initialize = gail_button_real_initialize;
-
-  widget_class->notify_gtk = gail_button_real_notify_gtk;
 }
 
 static void
@@ -211,6 +215,7 @@ gail_button_object_init (GailButton      *button)
   button->click_keybinding = NULL;
   button->action_queue = NULL;
   button->action_idle_handler = 0;
+  button->textutil = NULL;
 }
 
 AtkObject* 
@@ -256,7 +261,7 @@ gail_button_get_name (AtkObject *obj)
 
       g_return_val_if_fail (GTK_IS_BUTTON (widget), NULL);
 
-      child = get_label_from_button (widget);
+      child = get_label_from_button (widget, 0, FALSE);
       if (GTK_IS_LABEL (child))
         name = gtk_label_get_text (GTK_LABEL (child)); 
 
@@ -303,25 +308,34 @@ gail_button_real_initialize (AtkObject *obj,
                     G_CALLBACK (gail_button_released_leave_handler),
                     NULL);
 
-  button->textutil = gail_text_util_new ();
 
-  label = get_label_from_button (GTK_WIDGET (data));
+  label = get_label_from_button (GTK_WIDGET (data), 0, FALSE);
   if (GTK_IS_LABEL (label))
     {
-      label_text = gtk_label_get_text (GTK_LABEL (label));
-      gail_text_util_text_setup (button->textutil, label_text);
+      gail_button_init_textutil (button, label);
     }
 }
 
 static void
-gail_button_real_notify_gtk (GObject           *obj,
-                             GParamSpec        *pspec)
+gail_button_notify_label_gtk (GObject           *obj,
+                              GParamSpec        *pspec,
+                              gpointer           data)
 {
-  GtkWidget *widget = GTK_WIDGET (obj);
-  AtkObject* atk_obj = gtk_widget_get_accessible (widget);
+  AtkObject* atk_obj = ATK_OBJECT (data);
+  GtkLabel *label;
+  GailButton *gail_button;
 
   if (strcmp (pspec->name, "label") == 0)
     {
+      const gchar* label_text;
+
+      label = GTK_LABEL (obj);
+
+      label_text = gtk_label_get_text (label);
+
+      gail_button = GAIL_BUTTON (atk_obj);
+      gail_text_util_text_setup (gail_button->textutil, label_text);
+
       if (atk_obj->name == NULL)
       {
         /*
@@ -334,8 +348,21 @@ gail_button_real_notify_gtk (GObject           *obj,
        */
       g_signal_emit_by_name (atk_obj, "visible_data_changed");
     }
-  else
-    GAIL_WIDGET_CLASS (parent_class)->notify_gtk (obj, pspec);
+}
+
+static void
+gail_button_init_textutil (GailButton  *button,
+                           GtkWidget   *label)
+{
+  const gchar *label_text;
+
+  button->textutil = gail_text_util_new ();
+  label_text = gtk_label_get_text (GTK_LABEL (label));
+  gail_text_util_text_setup (button->textutil, label_text);
+  g_signal_connect (label,
+                    "notify",
+                    (GCallback) gail_button_notify_label_gtk,
+                    button);     
 }
 
 static void
@@ -513,7 +540,7 @@ gail_button_get_keybinding (AtkAction *action,
 
         g_return_val_if_fail (GTK_IS_BUTTON (widget), NULL);
 
-        child = get_label_from_button (widget);
+        child = get_label_from_button (widget, 0, FALSE);
         if (GTK_IS_LABEL (child))
           {
             guint key_val; 
@@ -613,14 +640,53 @@ gail_button_set_description (AtkAction      *action,
 static gint
 gail_button_get_n_children (AtkObject* obj)
 {
-  return 0;
+  GtkWidget *widget;
+  gint n_children;
+
+  g_return_val_if_fail (GAIL_IS_BUTTON (obj), 0);
+
+  widget = GTK_ACCESSIBLE (obj)->widget;
+  if (widget == NULL)
+    /*
+     * State is defunct
+     */
+    return 0;
+
+  n_children = get_n_labels_from_button (widget);
+  if (n_children <= 1)
+    n_children = 0;
+
+  return n_children;
 }
 
 static AtkObject*
 gail_button_ref_child (AtkObject *obj,
                        gint      i)
 {
-  return NULL;
+  GtkWidget *widget;
+  AtkObject *child;
+
+  g_return_val_if_fail (GAIL_IS_BUTTON (obj), NULL);
+
+  widget = GTK_ACCESSIBLE (obj)->widget;
+  if (widget == NULL)
+    /*
+     * State is defunct
+     */
+    return NULL;
+
+  child = NULL;
+  if (get_n_labels_from_button (widget) > 1)
+    {
+      widget = get_label_from_button (widget, i, TRUE);
+      if (widget)
+        {
+          child = gtk_widget_get_accessible (widget);
+          g_object_ref (child);
+        }
+    }
+
+  return child;
 }
 
 static AtkStateSet*
@@ -865,7 +931,7 @@ gail_button_get_text (AtkText *text,
 {
   GtkWidget *widget;
   GtkWidget  *label;
-
+  GailButton *button;
   const gchar *label_text;
 
   widget = GTK_ACCESSIBLE (text)->widget;
@@ -873,18 +939,22 @@ gail_button_get_text (AtkText *text,
     /* State is defunct */
     return NULL;
 
-  label = get_label_from_button (widget);
+  label = get_label_from_button (widget, 0, FALSE);
 
   if (!GTK_IS_LABEL (label))
     return NULL;
 
+  button = GAIL_BUTTON (text);
+  if (!button->textutil) 
+    gail_button_init_textutil (button, label);
+
   label_text = gtk_label_get_text (GTK_LABEL (label));
- 
+
   if (label_text == NULL)
     return NULL;
   else
   {
-    return gail_text_util_get_substring (GAIL_BUTTON(text)->textutil, 
+    return gail_text_util_get_substring (button->textutil, 
                                          start_pos, end_pos);
   }
 }
@@ -898,6 +968,7 @@ gail_button_get_text_before_offset (AtkText         *text,
 {
   GtkWidget *widget;
   GtkWidget *label;
+  GailButton *button;
   
   widget = GTK_ACCESSIBLE (text)->widget;
   
@@ -906,12 +977,16 @@ gail_button_get_text_before_offset (AtkText         *text,
     return NULL;
   
   /* Get label */
-  label = get_label_from_button (widget);
+  label = get_label_from_button (widget, 0, FALSE);
 
   if (!GTK_IS_LABEL(label))
     return NULL;
 
-  return gail_text_util_get_text (GAIL_BUTTON (text)->textutil,
+  button = GAIL_BUTTON (text);
+  if (!button->textutil)
+    gail_button_init_textutil (button, label);
+
+  return gail_text_util_get_text (button->textutil,
                            gtk_label_get_layout (GTK_LABEL (label)), GAIL_BEFORE_OFFSET, 
                            boundary_type, offset, start_offset, end_offset); 
 }
@@ -925,6 +1000,7 @@ gail_button_get_text_at_offset (AtkText         *text,
 {
   GtkWidget *widget;
   GtkWidget *label;
+  GailButton *button;
  
   widget = GTK_ACCESSIBLE (text)->widget;
   
@@ -933,12 +1009,16 @@ gail_button_get_text_at_offset (AtkText         *text,
     return NULL;
   
   /* Get label */
-  label = get_label_from_button (widget);
+  label = get_label_from_button (widget, 0, FALSE);
 
   if (!GTK_IS_LABEL(label))
     return NULL;
 
-  return gail_text_util_get_text (GAIL_BUTTON (text)->textutil,
+  button = GAIL_BUTTON (text);
+  if (!button->textutil)
+    gail_button_init_textutil (button, label);
+
+  return gail_text_util_get_text (button->textutil,
                               gtk_label_get_layout (GTK_LABEL (label)), GAIL_AT_OFFSET, 
                               boundary_type, offset, start_offset, end_offset);
 }
@@ -952,6 +1032,7 @@ gail_button_get_text_after_offset (AtkText         *text,
 {
   GtkWidget *widget;
   GtkWidget *label;
+  GailButton *button;
 
   widget = GTK_ACCESSIBLE (text)->widget;
   
@@ -962,12 +1043,16 @@ gail_button_get_text_after_offset (AtkText         *text,
   }
   
   /* Get label */
-  label = get_label_from_button (widget);
+  label = get_label_from_button (widget, 0, FALSE);
 
   if (!GTK_IS_LABEL(label))
     return NULL;
 
-  return gail_text_util_get_text (GAIL_BUTTON (text)->textutil,
+  button = GAIL_BUTTON (text);
+  if (!button->textutil)
+    gail_button_init_textutil (button, label);
+
+  return gail_text_util_get_text (button->textutil,
                            gtk_label_get_layout (GTK_LABEL (label)), GAIL_AFTER_OFFSET, 
                            boundary_type, offset, start_offset, end_offset);
 }
@@ -983,7 +1068,7 @@ gail_button_get_character_count (AtkText *text)
     /* State is defunct */
     return 0;
 
-  label = get_label_from_button (widget);
+  label = get_label_from_button (widget, 0, FALSE);
 
   if (!GTK_IS_LABEL(label))
     return 0;
@@ -1012,7 +1097,7 @@ gail_button_get_character_extents (AtkText      *text,
     /* State is defunct */
     return;
 
-  label = get_label_from_button (widget);
+  label = get_label_from_button (widget, 0, FALSE);
 
   if (!GTK_IS_LABEL(label))
     return;
@@ -1041,7 +1126,7 @@ gail_button_get_offset_at_point (AtkText      *text,
   if (widget == NULL)
     /* State is defunct */
     return -1;
-  label = get_label_from_button (widget);
+  label = get_label_from_button (widget, 0, FALSE);
 
   if (!GTK_IS_LABEL(label))
     return -1;
@@ -1080,7 +1165,7 @@ gail_button_get_run_attributes (AtkText        *text,
     /* State is defunct */
     return NULL;
 
-  label = get_label_from_button (widget);
+  label = get_label_from_button (widget, 0, FALSE);
 
   if (!GTK_IS_LABEL(label))
     return NULL;
@@ -1122,7 +1207,7 @@ gail_button_get_default_attributes (AtkText        *text)
     /* State is defunct */
     return NULL;
 
-  label = get_label_from_button (widget);
+  label = get_label_from_button (widget, 0, FALSE);
 
   if (!GTK_IS_LABEL(label))
     return NULL;
@@ -1147,7 +1232,7 @@ gail_button_get_character_at_offset (AtkText	         *text,
     /* State is defunct */
     return '\0';
 
-  label = get_label_from_button (widget);
+  label = get_label_from_button (widget, 0, FALSE);
 
   if (!GTK_IS_LABEL(label))
     return '\0';
@@ -1177,13 +1262,108 @@ gail_button_finalize (GObject            *object)
     {
       g_queue_free (button->action_queue);
     }
+  if (button->textutil)
+    {
+      g_object_unref (button->textutil);
+    }
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
 static GtkWidget*
-get_label_from_button (GtkWidget *button)
+get_label_from_button (GtkWidget *button,
+                       gint      index,
+                       gboolean  allow_many)
 {
   GtkWidget *child;
+  gint n_labels;
+
+  if (index > 0 && !allow_many)
+    g_warning ("Inconsistent values passed to get_label_from_button");
+
+  n_labels = 0;
+  child = gtk_bin_get_child (GTK_BIN (button));
+  if (GTK_IS_ALIGNMENT (child))
+    child = gtk_bin_get_child (GTK_BIN (child));
+
+  while (GTK_IS_CONTAINER (child))
+    {
+      /*
+       * Child is not a label.  Search for a label in the container's children.
+       */
+      GList *children, *tmp_list;
+ 
+      children = gtk_container_get_children (GTK_CONTAINER (child));
+
+      child = NULL;
+      for (tmp_list = children; tmp_list != NULL; tmp_list = tmp_list->next) 
+	{
+	  if (GTK_IS_LABEL (tmp_list->data))
+	    {
+              if (!allow_many)
+                {
+                  if (child)
+                    {
+                      child = NULL;
+                      break;
+                    }
+	          child = GTK_WIDGET (tmp_list->data);
+                }
+              else
+                {
+                  if (n_labels == index)
+                    {
+	              child = GTK_WIDGET (tmp_list->data);
+                      break;
+                    }
+                  n_labels++;
+	        }
+	    }
+           /*
+            * Label for button which are GtkTreeView column headers are in a 
+            * GtkHBox in a GtkAlignment.
+            */
+          if (GTK_IS_ALIGNMENT (tmp_list->data))
+            {
+              GtkWidget *widget;
+
+              widget = gtk_bin_get_child (GTK_BIN (tmp_list->data));
+              if (GTK_IS_LABEL (widget))
+                {
+                  if (!allow_many)
+                    {
+                      if (child)
+                        {
+                          child = NULL;
+                          break;
+                        }
+	              child = GTK_WIDGET (tmp_list->data);
+                    }
+                  else
+                    {
+                      if (n_labels == index)
+                        {
+	                  child = GTK_WIDGET (tmp_list->data);
+                          break;
+                        }
+                      n_labels++;
+	            }
+	        }
+	    }
+	}
+
+      g_list_free (children);
+    }
+
+  return child;
+}
+
+static gint
+get_n_labels_from_button (GtkWidget *button)
+{
+  GtkWidget *child;
+  gint n_labels;
+
+  n_labels = 0;
 
   child = gtk_bin_get_child (GTK_BIN (button));
   if (GTK_IS_ALIGNMENT (child))
@@ -1201,10 +1381,8 @@ get_label_from_button (GtkWidget *button)
       for (tmp_list = children; tmp_list != NULL; tmp_list = tmp_list->next) 
 	{
 	  if (GTK_IS_LABEL (tmp_list->data))
-	    {
-	      child = GTK_WIDGET (tmp_list->data);
-	      break;
-	    }  
+            n_labels++;
+
            /*
             * Label for button which are GtkreeView column headers are in a 
             * GtkHBox in a GtkAlignment.
@@ -1212,13 +1390,14 @@ get_label_from_button (GtkWidget *button)
           if (GTK_IS_ALIGNMENT (tmp_list->data))
             {
               child = gtk_bin_get_child (GTK_BIN (tmp_list->data));
-              break;
+	      if (GTK_IS_LABEL (child))
+                n_labels++;
 	    }
 	}
 
       g_list_free (children);
     }
-  return child;
+  return n_labels;
 }
 
 static void
