@@ -61,6 +61,7 @@ static G_CONST_RETURN gchar *
 				                   gint                index);
 static gboolean     gail_cell_action_do_action    (AtkAction           *action,
 			                           gint                index);
+static gboolean     idle_do_action                (gpointer            data);
 
 static void         atk_component_interface_init  (AtkComponentIface   *iface);
 static void         gail_cell_get_extents         (AtkComponent        *component,
@@ -69,6 +70,7 @@ static void         gail_cell_get_extents         (AtkComponent        *componen
                                                    gint                *width,
                                                    gint                *height,
                                                    AtkCoordType        coord_type);
+static gboolean     gail_cell_grab_focus         (AtkComponent        *component);
 
 static gpointer parent_class = NULL;
 
@@ -116,7 +118,7 @@ gail_cell_class_init (GailCellClass *klass)
   AtkObjectClass *class = ATK_OBJECT_CLASS (klass);
   GObjectClass *g_object_class = G_OBJECT_CLASS (klass);
 
-  parent_class = g_type_class_ref (ATK_TYPE_OBJECT);
+  parent_class = g_type_class_peek_parent (klass);
   g_object_class->finalize = gail_cell_object_finalize;
 
   class->get_index_in_parent = gail_cell_get_index_in_parent;
@@ -173,6 +175,11 @@ gail_cell_object_finalize (GObject *obj)
   if (cell->state_set)
     g_object_unref (cell->state_set);
   g_list_free (cell->action_list);
+  if (cell->action_idle_handler)
+    {
+      gtk_idle_remove (cell->action_idle_handler);
+      cell->action_idle_handler = 0;
+    }
   G_OBJECT_CLASS (parent_class)->finalize (obj);
 }
 
@@ -343,7 +350,9 @@ gail_cell_remove_action (GailCell *cell,
 
   g_return_val_if_fail (GAIL_IS_CELL (cell), FALSE);
   list_node = g_list_nth (cell->action_list, action_index);
-  g_return_val_if_fail ((list_node != NULL), FALSE);
+  if (!list_node)
+    return FALSE;
+  g_return_val_if_fail (list_node->data != NULL, FALSE);
   _gail_cell_destroy_action_info (list_node->data, NULL);
   cell->action_list = g_list_remove_link (cell->action_list, list_node);
   return TRUE;
@@ -383,7 +392,8 @@ _gail_cell_get_action_info (GailCell *cell,
   if (cell->action_list == NULL)
     return NULL;
   list_node = g_list_nth (cell->action_list, index);
-  g_return_val_if_fail (list_node, NULL);
+  if (!list_node)
+    return NULL;
   return (ActionInfo *) (list_node->data);
 }
 
@@ -468,7 +478,23 @@ gail_cell_action_do_action (AtkAction *action,
   if (info == NULL)
     return FALSE;
   g_return_val_if_fail (info->do_action_func, FALSE);
-  return info->do_action_func (cell);
+  if (cell->action_idle_handler)
+    return FALSE;
+  cell->action_func = info->do_action_func;
+  cell->action_idle_handler = gtk_idle_add (idle_do_action, cell);
+  return TRUE;
+}
+
+static gboolean
+idle_do_action (gpointer data)
+{
+  GailCell *cell;
+
+  cell = GAIL_CELL (data);
+  cell->action_idle_handler = 0;
+  cell->action_func (cell);
+
+  return FALSE;
 }
 
 static void
@@ -477,6 +503,7 @@ atk_component_interface_init (AtkComponentIface *iface)
   g_return_if_fail (iface != NULL);
 
   iface->get_extents = gail_cell_get_extents;
+  iface->grab_focus = gail_cell_grab_focus;
 }
 
 static void 
@@ -500,3 +527,20 @@ gail_cell_get_extents (AtkComponent *component,
   gail_cell_parent_get_cell_extents (GAIL_CELL_PARENT (cell_parent), 
                                    gailcell, x, y, width, height, coord_type);
 }
+
+static gboolean 
+gail_cell_grab_focus (AtkComponent *component)
+{
+  GailCell *gailcell;
+  AtkObject *cell_parent;
+
+  g_return_val_if_fail (GAIL_IS_CELL (component), FALSE);
+
+  gailcell = GAIL_CELL (component);
+
+  cell_parent = gtk_widget_get_accessible (gailcell->widget);
+
+  g_return_val_if_fail (GAIL_IS_CELL_PARENT (cell_parent), FALSE);
+  return gail_cell_parent_grab_focus (GAIL_CELL_PARENT (cell_parent), 
+                                      gailcell);
+}
