@@ -21,12 +21,14 @@
 #include <gtk/gtk.h>
 #include "gailentry.h"
 #include "gailcombo.h"
-#include <util/gailmisc.h>
+#include <libgail-util/gailmisc.h>
 
 static void       gail_entry_class_init            (GailEntryClass       *klass);
 static void       gail_entry_object_init           (GailEntry            *entry);
 static void	  gail_entry_real_initialize       (AtkObject            *obj,
                                                     gpointer             data);
+static void       text_setup                       (GailEntry            *entry,
+                                                    GtkEntry             *gtk_entry);
 static void	  gail_entry_real_notify_gtk	   (GObject		 *obj,
                                                     GParamSpec		 *pspec);
 static void       gail_entry_finalize              (GObject              *object);
@@ -133,8 +135,22 @@ static void	  _gail_entry_delete_text_cb	   (GtkEntry		 *entry,
                                                     gint		 arg1,
                                                     gint		 arg2);
 static void	  _gail_entry_changed_cb           (GtkEntry		 *entry);
-static gboolean    check_for_selection_change       (GailEntry           *entry,
-                                                     GtkEntry            *gail_entry);
+static gboolean   check_for_selection_change       (GailEntry            *entry,
+                                                    GtkEntry             *gtk_entry);
+
+static void                  atk_action_interface_init   (AtkActionIface  *iface);
+
+static gboolean              gail_entry_do_action        (AtkAction       *action,
+                                                          gint            i);
+static gboolean              idle_do_action              (gpointer        data);
+static gint                  gail_entry_get_n_actions    (AtkAction       *action);
+static G_CONST_RETURN gchar* gail_entry_get_description  (AtkAction       *action,
+                                                          gint            i);
+static G_CONST_RETURN gchar* gail_entry_action_get_name  (AtkAction       *action,
+                                                          gint            i);
+static gboolean              gail_entry_set_description  (AtkAction       *action,
+                                                          gint            i,
+                                                          const gchar     *desc);
 
 static GailWidgetClass *parent_class = NULL;
 
@@ -181,13 +197,21 @@ gail_entry_get_type (void)
         NULL
       };
 
+      static const GInterfaceInfo atk_action_info =
+      {
+        (GInterfaceInitFunc) atk_action_interface_init,
+        (GInterfaceFinalizeFunc) NULL,
+        NULL
+      };
+
       type = g_type_register_static (GAIL_TYPE_WIDGET,
                                      "GailEntry", &tinfo, 0);
       g_type_add_interface_static (type, ATK_TYPE_EDITABLE_TEXT,
                                    &atk_editable_text_info);
       g_type_add_interface_static (type, ATK_TYPE_TEXT,
                                    &atk_text_info);
-
+      g_type_add_interface_static (type, ATK_TYPE_ACTION,
+                                   &atk_action_info);
     }
   return type;
 }
@@ -219,6 +243,7 @@ gail_entry_object_init (GailEntry *entry)
   entry->signal_name = NULL;
   entry->cursor_position = 0;
   entry->selection_bound = 0;
+  entry->activate_description = NULL;
 }
 
 static void
@@ -227,7 +252,6 @@ gail_entry_real_initialize (AtkObject *obj,
 {
   GtkEntry *entry;
   GailEntry *gail_entry;
-  const gchar *entry_text;
 
   ATK_OBJECT_CLASS (parent_class)->initialize (obj, data);
 
@@ -237,8 +261,9 @@ gail_entry_real_initialize (AtkObject *obj,
   g_assert (GTK_IS_ENTRY (data));
 
   entry = GTK_ENTRY (data);
-  entry_text = gtk_entry_get_text (entry);
-  gail_text_util_text_setup (gail_entry->textutil, entry_text);
+  text_setup (gail_entry, entry);
+  gail_entry->cursor_position = entry->current_pos;
+  gail_entry->selection_bound = entry->selection_bound;
 
   /* Set up signal callbacks */
   g_signal_connect (data, "insert-text",
@@ -253,13 +278,26 @@ static void
 gail_entry_real_notify_gtk (GObject		*obj,
                             GParamSpec		*pspec)
 {
-  GtkWidget *widget = GTK_WIDGET (obj);
-  AtkObject* atk_obj = gtk_widget_get_accessible (widget);
-  GtkEntry* gtk_entry = GTK_ENTRY (widget);
-  GailEntry* entry = GAIL_ENTRY (atk_obj);
+  GtkWidget *widget;
+  AtkObject* atk_obj;
+  GtkEntry* gtk_entry;
+  GailEntry* entry;
+
+  widget = GTK_WIDGET (obj);
+  atk_obj = gtk_widget_get_accessible (widget);
+  gtk_entry = GTK_ENTRY (widget);
+  entry = GAIL_ENTRY (atk_obj);
 
   if (strcmp (pspec->name, "cursor-position") == 0)
     {
+      if (entry->signal_name)
+        {
+          g_signal_emit_by_name (atk_obj, 
+                                 entry->signal_name,
+                                 entry->position,
+                                 entry->length);
+          entry->signal_name = NULL;
+        }
       if (check_for_selection_change (entry, gtk_entry))
         g_signal_emit_by_name (atk_obj, "text_selection_changed");
       /*
@@ -270,6 +308,14 @@ gail_entry_real_notify_gtk (GObject		*obj,
     }
   else if (strcmp (pspec->name, "selection-bound") == 0)
     {
+      if (entry->signal_name)
+        {
+          g_signal_emit_by_name (atk_obj, 
+                                 entry->signal_name,
+                                 entry->position,
+                                 entry->length);
+          entry->signal_name = NULL;
+        }
       if (check_for_selection_change (entry, gtk_entry))
         g_signal_emit_by_name (atk_obj, "text_selection_changed");
     }
@@ -282,16 +328,68 @@ gail_entry_real_notify_gtk (GObject		*obj,
       atk_object_notify_state_change (atk_obj, ATK_STATE_EDITABLE,
                                                g_value_get_boolean (&value));
     }
+  else if (strcmp (pspec->name, "visibility") == 0)
+    {
+      gboolean visibility;
+      AtkRole new_role;
+
+      text_setup (entry, gtk_entry);
+      visibility = gtk_entry_get_visibility (gtk_entry);
+      new_role = visibility ? ATK_ROLE_TEXT : ATK_ROLE_PASSWORD_TEXT;
+      atk_object_set_role (atk_obj, new_role);
+    }
+  else if (strcmp (pspec->name, "invisible-char") == 0)
+    {
+      text_setup (entry, gtk_entry);
+    }
   else
     parent_class->notify_gtk (obj, pspec);
 }
 
+static void
+text_setup (GailEntry *entry,
+            GtkEntry  *gtk_entry)
+{
+  if (gtk_entry_get_visibility (gtk_entry))
+    {
+      gail_text_util_text_setup (entry->textutil, gtk_entry_get_text (gtk_entry));
+    }
+  else
+    {
+      gunichar invisible_char;
+      GString *tmp_string = g_string_new (NULL);
+      gint ch_len; 
+      gchar buf[7];
+      gint i;
+
+      invisible_char = gtk_entry_get_invisible_char (gtk_entry);
+      if (invisible_char == 0)
+        invisible_char = ' ';
+
+      ch_len = g_unichar_to_utf8 (invisible_char, buf);
+      for (i = 0; i < gtk_entry->text_length; i++)
+        {
+          g_string_append_len (tmp_string, buf, ch_len);
+        }
+
+      gail_text_util_text_setup (entry->textutil, tmp_string->str);
+      g_string_free (tmp_string, TRUE);
+
+    } 
+}
+
 static void
 gail_entry_finalize (GObject            *object)
 {
   GailEntry *entry = GAIL_ENTRY (object);
 
   g_object_unref (entry->textutil);
+  g_free (entry->activate_description);
+  if (entry->action_idle_handler)
+    {
+      gtk_idle_remove (entry->action_idle_handler);
+      entry->action_idle_handler = 0;
+    }
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -388,7 +486,6 @@ gail_entry_get_text (AtkText *text,
                      gint    start_pos,
                      gint    end_pos)
 {
-  GtkEntry *entry;
   GtkWidget *widget;
 
   widget = GTK_ACCESSIBLE (text)->widget;
@@ -396,9 +493,7 @@ gail_entry_get_text (AtkText *text,
     /* State is defunct */
     return NULL;
 
-  entry = GTK_ENTRY (widget);
-  
-  return gtk_editable_get_chars (GTK_EDITABLE (entry), start_pos, end_pos);
+  return gail_text_util_get_substring (GAIL_ENTRY (text)->textutil, start_pos, end_pos);
 }
 
 static gchar*
@@ -628,7 +723,12 @@ gail_entry_get_offset_at_point (AtkText *text,
   index = gail_misc_get_index_at_point_in_layout (widget, 
                gtk_entry_get_layout(entry), x_layout, y_layout, x, y, coords);
   if (index == -1)
-    return -1;  
+    {
+      if (coords == ATK_XY_SCREEN || coords == ATK_XY_WINDOW)
+        return g_utf8_strlen (entry->text, -1);
+
+      return index;  
+    }
   else
     {
       cursor_index = g_utf8_offset_to_pointer (entry->text, 
@@ -645,22 +745,6 @@ gail_entry_get_offset_at_point (AtkText *text,
     }
 }
 
-/* atkeditable.h */
-
-static void
-atk_editable_text_interface_init (AtkEditableTextIface *iface)
-{
-  g_return_if_fail (iface != NULL);
-
-  iface->set_text_contents = gail_entry_set_text_contents;
-  iface->insert_text = gail_entry_insert_text;
-  iface->copy_text = gail_entry_copy_text;
-  iface->cut_text = gail_entry_cut_text;
-  iface->delete_text = gail_entry_delete_text;
-  iface->paste_text = gail_entry_paste_text;
-  iface->set_run_attributes = NULL;;
-}
-
 static gint
 gail_entry_get_n_selections (AtkText              *text)
 {
@@ -812,12 +896,27 @@ gail_entry_set_selection (AtkText *text,
     return FALSE;
 }
 
+static void
+atk_editable_text_interface_init (AtkEditableTextIface *iface)
+{
+  g_return_if_fail (iface != NULL);
+
+  iface->set_text_contents = gail_entry_set_text_contents;
+  iface->insert_text = gail_entry_insert_text;
+  iface->copy_text = gail_entry_copy_text;
+  iface->cut_text = gail_entry_cut_text;
+  iface->delete_text = gail_entry_delete_text;
+  iface->paste_text = gail_entry_paste_text;
+  iface->set_run_attributes = NULL;
+}
+
 static void
 gail_entry_set_text_contents (AtkEditableText *text,
                               const gchar     *string)
 {
   GtkEntry *entry;
   GtkWidget *widget;
+  GtkEditable *editable;
 
   widget = GTK_ACCESSIBLE (text)->widget;
   if (widget == NULL)
@@ -825,6 +924,9 @@ gail_entry_set_text_contents (AtkEditableText *text,
     return;
 
   entry = GTK_ENTRY (widget);
+  editable = GTK_EDITABLE (entry);
+  if (!gtk_editable_get_editable (editable))
+    return;
 
   gtk_entry_set_text (entry, string);
 }
@@ -837,6 +939,7 @@ gail_entry_insert_text (AtkEditableText *text,
 {
   GtkEntry *entry;
   GtkWidget *widget;
+  GtkEditable *editable;
 
   widget = GTK_ACCESSIBLE (text)->widget;
   if (widget == NULL)
@@ -844,8 +947,11 @@ gail_entry_insert_text (AtkEditableText *text,
     return;
 
   entry = GTK_ENTRY (widget);
+  editable = GTK_EDITABLE (entry);
+  if (!gtk_editable_get_editable (editable))
+    return;
 
-  gtk_editable_insert_text (GTK_EDITABLE (entry), string, length, position);
+  gtk_editable_insert_text (editable, string, length, position);
 }
 
 static void
@@ -855,6 +961,7 @@ gail_entry_copy_text   (AtkEditableText *text,
 {
   GtkEntry *entry;
   GtkWidget *widget;
+  GtkEditable *editable;
   gchar *str;
 
   widget = GTK_ACCESSIBLE (text)->widget;
@@ -863,8 +970,9 @@ gail_entry_copy_text   (AtkEditableText *text,
     return;
 
   entry = GTK_ENTRY (widget);
-  str = gtk_editable_get_chars (GTK_EDITABLE (entry), start_pos, end_pos);
-  gtk_clipboard_set_text(gtk_clipboard_get(GDK_NONE), str, -1);
+  editable = GTK_EDITABLE (entry);
+  str = gtk_editable_get_chars (editable, start_pos, end_pos);
+  gtk_clipboard_set_text (gtk_clipboard_get (GDK_NONE), str, -1);
 }
 
 static void
@@ -874,6 +982,7 @@ gail_entry_cut_text (AtkEditableText *text,
 {
   GtkEntry *entry;
   GtkWidget *widget;
+  GtkEditable *editable;
   gchar *str;
 
   widget = GTK_ACCESSIBLE (text)->widget;
@@ -882,9 +991,12 @@ gail_entry_cut_text (AtkEditableText *text,
     return;
 
   entry = GTK_ENTRY (widget);
-  str = gtk_editable_get_chars (GTK_EDITABLE (entry), start_pos, end_pos);
+  editable = GTK_EDITABLE (entry);
+  if (!gtk_editable_get_editable (editable))
+    return;
+  str = gtk_editable_get_chars (editable, start_pos, end_pos);
   gtk_clipboard_set_text (gtk_clipboard_get (GDK_NONE), str, -1);
-  gtk_editable_delete_text (GTK_EDITABLE (entry), start_pos, end_pos);
+  gtk_editable_delete_text (editable, start_pos, end_pos);
 }
 
 static void
@@ -894,6 +1006,7 @@ gail_entry_delete_text (AtkEditableText *text,
 {
   GtkEntry *entry;
   GtkWidget *widget;
+  GtkEditable *editable;
 
   widget = GTK_ACCESSIBLE (text)->widget;
   if (widget == NULL)
@@ -901,8 +1014,11 @@ gail_entry_delete_text (AtkEditableText *text,
     return;
 
   entry = GTK_ENTRY (widget);
+  editable = GTK_EDITABLE (entry);
+  if (!gtk_editable_get_editable (editable))
+    return;
 
-  gtk_editable_delete_text (GTK_EDITABLE (entry), start_pos, end_pos);
+  gtk_editable_delete_text (editable, start_pos, end_pos);
 }
 
 static void
@@ -910,6 +1026,7 @@ gail_entry_paste_text (AtkEditableText *text,
                        gint            position)
 {
   GtkWidget *widget;
+  GtkEditable *editable;
   GailEntryPaste paste_struct;
 
   widget = GTK_ACCESSIBLE (text)->widget;
@@ -917,6 +1034,9 @@ gail_entry_paste_text (AtkEditableText *text,
     /* State is defunct */
     return;
 
+  editable = GTK_EDITABLE (widget);
+  if (!gtk_editable_get_editable (editable))
+    return;
   paste_struct.entry = GTK_ENTRY (widget);
   paste_struct.position = position;
 
@@ -962,30 +1082,40 @@ static void _gail_entry_insert_text_cb (GtkEntry *entry,
       gail_entry->length = arg2;
     }
   /*
-   * The signal will be emitted when the changed signal is received
+   * The signal will be emitted when the cursor position is updated.
    */
 }
 
 static gunichar 
-gail_entry_get_character_at_offset (AtkText	         *text,
-                                    gint	         offset)
+gail_entry_get_character_at_offset (AtkText *text,
+                                    gint     offset)
 {
   GtkWidget *widget;
-  GtkEntry *entry;
+  GailEntry *entry;
   gchar *string;
   gchar *index;
+  gunichar unichar;
 
   widget = GTK_ACCESSIBLE (text)->widget;
   if (widget == NULL)
     /* State is defunct */
     return '\0';
 
-  entry = GTK_ENTRY (widget);
-  string = gtk_editable_get_chars (GTK_EDITABLE (entry), 0, -1);
-  index = g_utf8_offset_to_pointer (string, offset);
-  g_free(string);
+  entry = GAIL_ENTRY (text);
+  string = gail_text_util_get_substring (entry->textutil, 0, -1);
+  if (offset >= g_utf8_strlen (string, -1))
+    {
+      unichar = '\0';
+    }
+  else
+    {
+      index = g_utf8_offset_to_pointer (string, offset);
 
-  return g_utf8_get_char(index);
+      unichar = g_utf8_get_char(index);
+    }
+
+  g_free(string);
+  return unichar;
 }
 
 /* Note arg1 returns the start of the delete range, arg2 returns the
@@ -999,6 +1129,12 @@ static void _gail_entry_delete_text_cb (GtkEntry *entry,
 
   accessible = gtk_widget_get_accessible (GTK_WIDGET (entry));
 
+  /*
+   * Zero length text deleted so ignore
+   */
+  if (arg2 - arg1 == 0)
+    return;
+
   g_signal_emit_by_name (accessible, "text_changed::delete", arg1, arg2 - arg1);
 }
 
@@ -1012,36 +1148,23 @@ _gail_entry_changed_cb (GtkEntry *entry)
 
   gail_entry = GAIL_ENTRY (accessible);
 
-  if (gail_entry->signal_name)
-    {
-      g_signal_emit_by_name (accessible, 
-                             gail_entry->signal_name,
-                             gail_entry->position,
-                             gail_entry->length);
-      gail_entry->signal_name = NULL;
-    }
-  /*
-   * Update the text in the GtkTextBuffer
-   */
-  gail_text_util_text_setup (gail_entry->textutil, gtk_entry_get_text(entry));
+  text_setup (gail_entry, entry);
 }
 
 static gboolean 
 check_for_selection_change (GailEntry   *entry,
-                            GtkEntry    *gail_entry)
+                            GtkEntry    *gtk_entry)
 {
-  gint pos;
   gboolean ret_val = FALSE;
  
-  pos = gtk_editable_get_position (GTK_EDITABLE (gail_entry));
-  if (gail_entry->current_pos != gail_entry->selection_bound)
+  if (gtk_entry->current_pos != gtk_entry->selection_bound)
     {
-      if (gail_entry->current_pos != entry->cursor_position ||
-          gail_entry->selection_bound != entry->selection_bound)
+      if (gtk_entry->current_pos != entry->cursor_position ||
+          gtk_entry->selection_bound != entry->selection_bound)
         /*
          * This check is here as this function can be called
-         * for notification of selection_bound or current_pos.
-         * The values of current_pos  selection_bound may be the same 
+         * for notification of selection_bound and current_pos.
+         * The values of current_pos and selection_bound may be the same 
          * for both notifications and we only want to generate one
          * text_selection_changed signal.
          */
@@ -1052,8 +1175,149 @@ check_for_selection_change (GailEntry   *entry,
       /* We had a selection */
       ret_val = (entry->cursor_position != entry->selection_bound);
     }
-  entry->cursor_position = gail_entry->current_pos;
-  entry->selection_bound = gail_entry->selection_bound;
+  entry->cursor_position = gtk_entry->current_pos;
+  entry->selection_bound = gtk_entry->selection_bound;
 
   return ret_val;
 }
+
+static void
+atk_action_interface_init (AtkActionIface *iface)
+{
+  g_return_if_fail (iface != NULL);
+
+  iface->do_action = gail_entry_do_action;
+  iface->get_n_actions = gail_entry_get_n_actions;
+  iface->get_description = gail_entry_get_description;
+  iface->get_name = gail_entry_action_get_name;
+  iface->set_description = gail_entry_set_description;
+}
+
+static gboolean
+gail_entry_do_action (AtkAction *action,
+                      gint      i)
+{
+  GailEntry *entry;
+  GtkWidget *widget;
+  gboolean return_value = TRUE;
+
+  entry = GAIL_ENTRY (action);
+  widget = GTK_ACCESSIBLE (action)->widget;
+  if (widget == NULL)
+    /*
+     * State is defunct
+     */
+    return FALSE;
+
+  if (!GTK_WIDGET_SENSITIVE (widget) || !GTK_WIDGET_VISIBLE (widget))
+    return FALSE;
+
+  switch (i)
+    {
+    case 0:
+      if (entry->action_idle_handler)
+        return_value = FALSE;
+      else
+        entry->action_idle_handler = gtk_idle_add (idle_do_action, entry);
+      break;
+    default:
+      return_value = FALSE;
+      break;
+    }
+  return return_value; 
+}
+
+static gboolean
+idle_do_action (gpointer data)
+{
+  GailEntry *entry;
+  GtkWidget *widget;
+
+  entry = GAIL_ENTRY (data);
+  entry->action_idle_handler = 0;
+  widget = GTK_ACCESSIBLE (entry)->widget;
+  if (widget == NULL)
+    /*
+     * State is defunct
+     */
+    return FALSE;
+
+  if (!GTK_WIDGET_SENSITIVE (widget) || !GTK_WIDGET_VISIBLE (widget))
+    return FALSE;
+
+  gtk_widget_activate (widget);
+  return FALSE;
+}
+
+static gint
+gail_entry_get_n_actions (AtkAction *action)
+{
+  return 1;
+}
+
+static G_CONST_RETURN gchar*
+gail_entry_get_description (AtkAction *action,
+                            gint      i)
+{
+  GailEntry *entry;
+  G_CONST_RETURN gchar *return_value;
+
+  entry = GAIL_ENTRY (action);
+  switch (i)
+    {
+    case 0:
+      return_value = entry->activate_description;
+      break;
+    default:
+      return_value = NULL;
+      break;
+    }
+  return return_value; 
+}
+
+static G_CONST_RETURN gchar*
+gail_entry_action_get_name (AtkAction *action,
+                            gint      i)
+{
+  G_CONST_RETURN gchar *return_value;
+
+  switch (i)
+    {
+    case 0:
+      return_value = "activate";
+      break;
+    default:
+      return_value = NULL;
+      break;
+  }
+  return return_value; 
+}
+
+static gboolean
+gail_entry_set_description (AtkAction      *action,
+                            gint           i,
+                            const gchar    *desc)
+{
+  GailEntry *entry;
+  gchar **value;
+
+  entry = GAIL_ENTRY (action);
+  switch (i)
+    {
+    case 0:
+      value = &entry->activate_description;
+      break;
+    default:
+      value = NULL;
+      break;
+    }
+
+  if (value)
+    {
+      g_free (*value);
+      *value = g_strdup (desc);
+      return TRUE;
+    }
+  else
+    return FALSE;
+}
