@@ -19,7 +19,6 @@
 
 #include <gtk/gtk.h>
 #include "gailimage.h"
-#include "gailimagefactory.h"
 
 static void      gail_image_class_init         (GailImageClass *klass);
 static void      gail_image_object_init        (GailImage      *image);
@@ -28,6 +27,10 @@ static void      atk_image_interface_init      (AtkImageIface  *iface);
 
 static G_CONST_RETURN gchar *
                  gail_image_get_image_description (AtkImage     *image);
+static void	 gail_image_get_image_position    (AtkImage     *image,
+                                                   gint         *x,
+                                                   gint         *y,
+                                                   AtkCoordType coord_type);
 static void      gail_image_get_image_size     (AtkImage        *image,
                                                 gint            *width,
                                                 gint            *height);
@@ -115,6 +118,7 @@ atk_image_interface_init (AtkImageIface *iface)
   g_return_if_fail (iface != NULL);
 
   iface->get_image_description = gail_image_get_image_description;
+  iface->get_image_position = gail_image_get_image_position;
   iface->get_image_size = gail_image_get_image_size;
   iface->set_image_description = gail_image_set_image_description;
 }
@@ -127,6 +131,15 @@ gail_image_get_image_description (AtkImage     *image)
   return aimage->image_description;
 }
 
+static void
+gail_image_get_image_position (AtkImage     *image,
+                               gint         *x,
+                               gint         *y,
+                               AtkCoordType coord_type)
+{
+  atk_component_get_position (ATK_COMPONENT (image), x, y, coord_type);
+}
+
 static void
 gail_image_get_image_size (AtkImage *image, 
                            gint     *width,
