@@ -22,6 +22,8 @@
 
 static void      gail_notebook_page_class_init      (GailNotebookPageClass     *klass);
 
+static void                  gail_notebook_page_finalize       (GObject   *object);
+
 static G_CONST_RETURN gchar* gail_notebook_page_get_name       (AtkObject *accessible);
 static AtkObject*            gail_notebook_page_get_parent     (AtkObject *accessible);
 static gint                  gail_notebook_page_get_n_children (AtkObject *accessible);
@@ -90,6 +92,7 @@ gail_notebook_page_get_type (void)
 static void
 gail_notebook_page_class_init (GailNotebookPageClass *klass)
 {
+  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
   AtkObjectClass *class = ATK_OBJECT_CLASS (klass);
   
   parent_class = g_type_class_peek_parent (klass);
@@ -100,6 +103,8 @@ gail_notebook_page_class_init (GailNotebookPageClass *klass)
   class->ref_child = gail_notebook_page_ref_child;
   class->ref_state_set = gail_notebook_page_ref_state_set;
   class->get_index_in_parent = gail_notebook_page_get_index_in_parent;
+
+  gobject_class->finalize = gail_notebook_page_finalize;
 }
 
 AtkObject*
@@ -113,26 +118,42 @@ gail_notebook_page_new (GtkNotebook *notebook,
   
   g_return_val_if_fail (GTK_IS_NOTEBOOK (notebook), NULL);
 
+  child = gtk_notebook_get_nth_page (notebook, pagenum);
+
+  if (!child)
+    return NULL;
+
   object = g_object_new (GAIL_TYPE_NOTEBOOK_PAGE, NULL);
   g_return_val_if_fail (object != NULL, NULL);
 
   page = GAIL_NOTEBOOK_PAGE (object);
   page->notebook = notebook;
+  g_object_add_weak_pointer (G_OBJECT (page->notebook), (gpointer *)&page->notebook);
   page->index = pagenum;
   
   atk_object = ATK_OBJECT (page);
   atk_object->role = ATK_ROLE_PAGE_TAB;
   atk_object->layer = ATK_LAYER_WIDGET;
 
-  child = gtk_notebook_get_nth_page (notebook, pagenum);
-  g_return_val_if_fail (GTK_IS_WIDGET (child), NULL);
-
   atk_child = gtk_widget_get_accessible (child);
   atk_object_set_parent (atk_child, atk_object);
   
   return atk_object;
 }
 
+static void
+gail_notebook_page_finalize (GObject *object)
+{
+  GailNotebookPage *page = GAIL_NOTEBOOK_PAGE (object);
+
+  if (page->notebook)
+    g_object_remove_weak_pointer (G_OBJECT (page->notebook), (gpointer *)&page->notebook);
+
+  G_OBJECT_CLASS (parent_class)->finalize (object);
+
+}
+
+
 static G_CONST_RETURN gchar*
 gail_notebook_page_get_name (AtkObject *accessible)
 {
@@ -150,6 +171,8 @@ gail_notebook_page_get_name (AtkObject *accessible)
       page = GAIL_NOTEBOOK_PAGE (accessible);
 	
       notebook = page->notebook;
+      if (!notebook)
+        return NULL;
 
       child = gtk_notebook_get_nth_page (notebook, page->index);
       g_return_val_if_fail (GTK_IS_WIDGET (child), NULL);
@@ -175,6 +198,9 @@ gail_notebook_page_get_parent (AtkObject *accessible)
   
   page = GAIL_NOTEBOOK_PAGE (accessible);
 
+  if (!page->notebook)
+    return NULL;
+
   return gtk_widget_get_accessible (GTK_WIDGET (page->notebook));
 }
 
@@ -196,9 +222,13 @@ gail_notebook_page_ref_child (AtkObject *accessible,
   GailNotebookPage *page = NULL;
    
   g_return_val_if_fail (GAIL_IS_NOTEBOOK_PAGE (accessible), NULL);
-  g_return_val_if_fail (i == 0, NULL);
+  if (i != 0)
+    return NULL;
    
   page = GAIL_NOTEBOOK_PAGE (accessible);
+  if (!page->notebook)
+    return NULL;
+
   child = gtk_notebook_get_nth_page (page->notebook, page->index);
   g_return_val_if_fail (GTK_IS_WIDGET (child), NULL);
    
@@ -289,13 +319,21 @@ gail_notebook_page_get_extents (AtkComponent *component,
 
   if (!atk_label)
     {
+      AtkObject *child;
+
       *width = 0;
       *height = 0;
-      *x = G_MININT;
-      *y = G_MININT;
+
+      child = atk_object_ref_accessible_child (ATK_OBJECT (component), 0);
+      g_return_if_fail (child);
+
+      atk_component_get_position (ATK_COMPONENT (child), x, y, coord_type);
+    }
+  else
+    {
+      atk_component_get_extents (ATK_COMPONENT (atk_label), 
+                                 x, y, width, height, coord_type);
     }
-  atk_component_get_extents (ATK_COMPONENT (atk_label), 
-                             x, y, width, height, coord_type);
   return; 
 }
 
@@ -304,6 +342,9 @@ _gail_notebook_page_get_tab_label (GailNotebookPage *page)
 {
   GtkWidget *label, *child;
 
+  if (!page->notebook)
+    return NULL;
+
   child = gtk_notebook_get_nth_page (page->notebook, page->index);
   label = gtk_notebook_get_tab_label (page->notebook, child);
   if (label)
