@@ -104,7 +104,7 @@ gail_scrollbar_get_index_in_parent (AtkObject *accessible)
     return ATK_OBJECT_CLASS (parent_class)->get_index_in_parent (accessible);
 
   scrolled_window = GTK_SCROLLED_WINDOW (widget->parent);
-  children = gtk_container_children (GTK_CONTAINER (scrolled_window));
+  children = gtk_container_get_children (GTK_CONTAINER (scrolled_window));
   n_children = g_list_length (children);
   g_list_free (children);
 
