@@ -94,7 +94,7 @@ gail_toplevel_new (void)
   g_return_val_if_fail ((object != NULL), NULL);
 
   accessible = ATK_OBJECT (object);
-  accessible->role = ATK_ROLE_INVALID;
+  accessible->role = ATK_ROLE_APPLICATION;
   accessible->name = g_get_prgname();
 
   return accessible;
