@@ -237,10 +237,10 @@ gail_widget_connect_widget_destroyed (GtkAccessible *accessible)
 {
   if (accessible->widget)
     {
-      g_signal_connect (accessible->widget,
-                        "destroy",
-                        G_CALLBACK (gail_widget_destroyed),
-                        accessible);
+      g_signal_connect_after (accessible->widget,
+                              "destroy",
+                              G_CALLBACK (gail_widget_destroyed),
+                              accessible);
     }
 }
 
