@@ -1,5 +1,4 @@
 #include <string.h>
-#include <stdlib.h>
 #include "testtextlib.h"
 
 #define NUM_ROWS_TO_LOOP 30
@@ -114,11 +113,11 @@ static void choicecb (GtkWidget *widget, gpointer data)
 static void _check_table (AtkObject *in_obj)
 {
   AtkObject *obj = 0;
-  char *no_properties;
-  char *no_gui;
+  G_CONST_RETURN char *no_properties;
+  G_CONST_RETURN char *no_gui;
 
-  no_properties = getenv ("TEST_ACCESSIBLE_NO_PROPERTIES");
-  no_gui = getenv ("TEST_ACCESSIBLE_NO_GUI");
+  no_properties = g_getenv ("TEST_ACCESSIBLE_NO_PROPERTIES");
+  no_gui = g_getenv ("TEST_ACCESSIBLE_NO_GUI");
 
   if (no_properties != NULL)
     g_properties = FALSE;
@@ -327,7 +326,7 @@ void other_runtest(AtkObject *obj)
       /* Generic cell tests */
       /* Just test setting column headers once. */
 
-      if (tested_set_headers == FALSE)
+      if (!tested_set_headers)
       {
         tested_set_headers = TRUE;
    
@@ -634,7 +633,7 @@ _display_header_info(gchar *type, AtkObject *header_obj, gint header_num)
     else 
     {
       g_print ("%d: %s header is of type %s!\n", header_num,
-        type, get_role_name(role));
+        type, atk_role_get_name (role));
     }
   }
   else
@@ -645,7 +644,7 @@ _display_header_info(gchar *type, AtkObject *header_obj, gint header_num)
 
 static void _property_signal_connect (AtkObject *obj)
 {
-  if (g_properties == TRUE && obj != NULL)
+  if (g_properties && obj != NULL)
   {
     g_signal_connect_closure_by_id (obj,
     g_signal_lookup ("property_change", G_OBJECT_TYPE (obj)),
@@ -756,7 +755,7 @@ void test_choice_gui(AtkObject **obj)
   GtkWidget *button;
 
 
-  tc = (TestChoice *) malloc (sizeof(TestChoice));
+  tc = (TestChoice *) g_malloc (sizeof(TestChoice));
   
   window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   gtk_window_set_title(GTK_WINDOW(window), "Test to run");
@@ -884,7 +883,7 @@ nogui_ref_at_runtest (AtkObject *obj)
       /* Generic cell tests */
       /* Just test setting column headers once. */
 
-      if (tested_set_headers == FALSE)
+      if (!tested_set_headers)
       {
         tested_set_headers = TRUE;
 
